# -*- coding: utf-8 -*-
"""
ISE

@author: Oliver A. Gubler (oliver.gubler@hevs.ch)
@version: 1.2016
    * update to python 3
@version: 1.2014
"""

import os
import subprocess
import re


class ISE:
    '''Do all kind of tasks with ISE without the GUI'''
    def __init__(self,
                 exec_path,
                 work_path,
                 source,
                 ucf,
                 fpga='xc3s500e-5-fg320',
                 flash='xcf04s',
                 project_name='ISE'):
        self.exec_path = os.path.join(exec_path, 'ISE', 'bin', 'nt64')
        self.work_path = work_path
        self.fpga = fpga
        self.flash = flash
        self.project_name = project_name
        self.source = source
        self.ucf = ucf
        self.work_file = os.path.join(work_path, project_name)

    def generate_bitfile(self):
        wd = os.getcwd()
        os.chdir(self.work_path)
        # ---------------------------------------------------------------------
        # ### run synthesis, build, map, par, trace and bitgen
        subprocess.call([os.path.join(self.exec_path, 'xst'),
                         '-intstyle', 'ise',
                         '-ofn', self.project_name + '.syr',
                         '-ifn', self.generate_xst()])
        subprocess.call([os.path.join(self.exec_path, 'ngdbuild'),
                         '-intstyle', 'ise',
                         '-dd', '_ngo',
                         '-aul',
                         '-nt', 'timestamp',
                         '-uc', self.ucf,
                         '-p', self.fpga,
                         self.work_file + '.ngc',
                         self.work_file + '.ngd'])
        subprocess.call([os.path.join(self.exec_path, 'map'),
                         '-intstyle', 'ise',
                         '-p', self.fpga,
                         '-cm', 'area',
                         '-ir', 'off',
                         '-pr', 'off',
                         '-c', '100',
                         '-o', self.work_file + '_map.ncd',
                         self.work_file + '.ngd',
                         self.work_file + '.pcf'])
        subprocess.call([os.path.join(self.exec_path, 'par'),
                         '-w',
                         '-intstyle', 'ise',
                         '-ol', 'high',
                         '-t', '1',
                         self.work_file + '_map.ncd',
                         self.work_file + '.ncd',
                         self.work_file + '.pcf'])
        subprocess.call([os.path.join(self.exec_path, 'trce'),
                         '-intstyle', 'ise',
                         '-v', '3',
                         '-s', '5',
                         '-n', '3',
                         '-fastpaths',
                         '-xml', self.work_file + '.twx',
                         '-o', self.work_file + '.twr',
                         self.work_file + '.ncd',
                         self.work_file + '.pcf'])
        subprocess.call([os.path.join(self.exec_path, 'bitgen'),
                         '-intstyle', 'ise',
                         '-w',
                         '-g', 'DebugBitstream:No',
                         '-g', 'Binary:no',
                         '-g', 'CRC:Enable',
                         '-g', 'ConfigRate:1',
                         '-g', 'ProgPin:PullUp',
                         '-g', 'DonePin:PullUp',
                         '-g', 'TckPin:PullUp',
                         '-g', 'TdiPin:PullUp',
                         '-g', 'TdoPin:PullUp',
                         '-g', 'TmsPin:PullUp',
                         '-g', 'UnusedPin:PullDown',
                         '-g', 'UserID:0xFFFFFFFF',
                         '-g', 'DCMShutdown:Disable',
                         '-g', 'StartUpClk:CClk',
                         '-g', 'DONE_cycle:4',
                         '-g', 'GTS_cycle:5',
                         '-g', 'GWE_cycle:6',
                         '-g', 'LCK_cycle:NoWait',
                         '-g', 'Security:None',
                         '-g', 'DonePipe:No',
                         '-g', 'DriveDone:No',
                         self.work_file + '.ncd'])
        os.chdir(wd)
        return self.work_file + '.bit'

    def generate_prj(self):
        '''prj file contains list of all source files for synthesis '''
        prj_file_name = self.work_file + '.prj'
        prj_file_id = open(prj_file_name, 'w')
        prj_file_id.write('vhdl work "' + self.source + '"')
        prj_file_id.close()
        return prj_file_name

    def generate_xst(self):
        '''xst file contains parameters for synthesis and subsequent steps '''
        xst_path = os.path.join(self.work_path, 'xst')
        tmp_path = os.path.join(xst_path, 'projnav.tmp')
        try:
            os.makedirs(tmp_path)
        except:
            print('WARNING! Folder could not be created:\n    ' + tmp_path)
        xst_file_name = self.work_file + '.xst'
        xst_file_id = open(xst_file_name, 'w')
        xst_file_id.write(
            'set -tmpdir "' + xst_path + '"' + '\n' +
            'set -xsthdpdir "' + tmp_path + '"' + '\n' +
            'run' + '\n' +
            '-ifn ' + self.generate_prj() + '\n' +
            '-ifmt mixed' + '\n' +
            '-ofn ' + self.work_file + '\n' +
            '-ofmt NGC' + '\n' +
            '-p ' + self.fpga + '\n' +
            '-top ' + find_top(self.source) + '\n' +
            '-opt_mode Speed' + '\n' +
            '-opt_level 1' + '\n' +
            '-iuc NO' + '\n' +
            '-keep_hierarchy NO' + '\n' +
            '-netlist_hierarchy as_optimized' + '\n' +
            '-rtlview Yes' + '\n' +
            '-glob_opt AllClockNets' + '\n' +
            '-read_cores YES' + '\n' +
            '-write_timing_constraints NO' + '\n' +
            '-cross_clock_analysis NO' + '\n' +
            '-hierarchy_separator /' + '\n' +
            '-bus_delimiter <>' + '\n' +
            '-case maintain' + '\n' +
            '-slice_utilization_ratio 100' + '\n' +
            '-bram_utilization_ratio 100' + '\n' +
            '-verilog2001 YES' + '\n' +
            '-fsm_extract YES' + '\n' +
            '-fsm_encoding Auto' + '\n' +
            '-safe_implementation No' + '\n' +
            '-fsm_style lut' + '\n' +
            '-ram_extract Yes' + '\n' +
            '-ram_style Auto' + '\n' +
            '-rom_extract Yes' + '\n' +
            '-mux_style Auto' + '\n' +
            '-decoder_extract YES' + '\n' +
            '-priority_extract YES' + '\n' +
            '-shreg_extract YES' + '\n' +
            '-shift_extract YES' + '\n' +
            '-xor_collapse YES' + '\n' +
            '-rom_style Auto' + '\n' +
            '-auto_bram_packing NO' + '\n' +
            '-mux_extract YES' + '\n' +
            '-resource_sharing YES' + '\n' +
            '-async_to_sync NO' + '\n' +
            '-mult_style auto' + '\n' +
            '-iobuf YES' + '\n' +
            '-max_fanout 500' + '\n' +
            '-bufg 24' + '\n' +
            '-register_duplication YES' + '\n' +
            '-register_balancing No' + '\n' +
            '-slice_packing YES' + '\n' +
            '-optimize_primitives NO' + '\n' +
            '-use_clock_enable Yes' + '\n' +
            '-use_sync_set Yes' + '\n' +
            '-use_sync_reset Yes' + '\n' +
            '-iob auto' + '\n' +
            '-equivalent_register_removal YES' + '\n' +
            '-slice_utilization_ratio_maxmargin 5' + '\n'
            '-loop_iteration_limit ' + '{}'.format(9999))
        xst_file_id.close()
        return xst_file_name

    def generate_mcs(self):
        print('Generate mcs: ')
        mcs_file_name = self.work_file + '.mcs'
        bit_file_name = self.work_file + '.bit'
        print('{} from {}'.format(mcs_file_name, bit_file_name))
        subprocess.call([os.path.join(self.exec_path, 'promgen'),
                         '-w',
                         '-p', 'mcs',
                         '-c', 'FF',
                         '-o', mcs_file_name,
                         '-x', self.flash,
                         '-u', '00000000',
                         bit_file_name])
        return mcs_file_name

    def configure_hw(self, program_flash=False, program_fpga=False):
        '''Program flash and/or fpga with impact'''
        if not program_flash and not program_fpga:
            input('#'*80 + '\n' +
                  'Neither FPGA nor FLASH are programmed.' + '\n' +
                  'Press Enter to continue!')
            return
        # cmd files contains impact commands
        cmd_file_name = self.project_name + '.cmd'
        cmd_file_id = open(cmd_file_name, 'w')
        cmd_file_id.write('setMode -bscan' + '\n' +
                          'setCable -port auto' + '\n' +
                          'identify -inferir' + '\n' +
                          'identifyMPM' + '\n' +
                          'assignFile -p 1' +
                              ' -file ' + self.work_file + '.mcs' '\n' +
                          'assignFile -p 2' +
                              ' -file ' + self.work_file + '.bit' + '\n')
        cmd_file_id.close()
        cmd_file_id = open(cmd_file_name, 'a')
        if program_flash:
            cmd_file_id.write('program -p 1 -e -v' + '\n')
        if program_fpga:
            cmd_file_id.write('program -p 2' + '\n')
        cmd_file_id.write('quit')
        cmd_file_id.close()
        input('#'*80 + '\n' +
              'Connect board and press Enter to program the FPGA!')
        subprocess.call([os.path.join(self.exec_path, 'impact'),
                        '-batch', cmd_file_name])
        input('#'*80 + '\n' +
              'Programming done.' + '\n' +
              'Press Enter to continue!')


def find_top(source):
    '''Get entity name of last architecture (a.k.a. toplevel) '''
    source_file_id = open(source, 'r')
    for line in source_file_id:
        # remove all comments
        line = re.sub('--.*', '', line)
        # find any architecture definition
        line = re.search('architecture\s*\w*\s*of\s*\w*\s*is',
                         line,
                         re.IGNORECASE)
        # store entity of last found architecture
        if line:
            fpga_top = line.group().split(' ')[3]
    return fpga_top
