# -*- coding: utf-8 -*-
"""Scritp to generate Xilinx RAM block initialisation text files and AMBA
command files to show a text, loaded from a file, on the oscilioscope with the
splines functionality. Furthermore the configuration files for the FPGA and the
FLASH are generated and can be programmed.
The parameters of this script are loaded form text2spline.ini.
All generated files are copied to the given output folder for further usage.

TODO: add some meaningful log

Original Matlab script:
@author: François Corthay (francois.corthay@hevs.ch)
Transcription to python and addition of PAR and programming:
@author: Oliver A. Gubler (oliver.gubler@hevs.ch)

@version: 1.1
    * create temporary or export path if not existant
    * add update_rate to settings
@version: 1.0
"""
import csv
import configparser
import tempfile
import shutil
import os
wd = os.getcwd()
import re
import matplotlib.pyplot as plt
plt.ion()
import spline_functions as s_f
import ISE
# #############################################################################
# # parameters
# =============================================================================
# ## external parameters
# -----------------------------------------------------------------------------
# ### default values, will be overwritten by values specified in ini file
# ### If you change these values, update the ini file accordingly
config = configparser.RawConfigParser({
    'text_input_file': os.path.join(wd, 'students.txt'),
    'ram_x_file': os.path.join(wd, '..', 'ramXInit.txt'),
    'ram_y_file': os.path.join(wd, '..', 'ramYInit.txt'),
    'signal_bit_nb': 16,
    'generate_amba_files': 'false',
    'update_period': 5,
    'generate_bitfile': 'true',
    'program_fpga': 'true',
    'program_flash': 'true',
    'concat_file': os.path.join(wd, '..\\..\\Board\\concat\\textWriter.vhd'),
    'ucf_file': os.path.join(wd, '..\\..\\Board\\concat\\textWriter.ucf'),
    'ise_work_path': os.path.join(wd, '..\\..\\Board\\ise'),
    'ise_path': 'C:\\eda\\Xilinx\\12.1\\ISE_DS\\',
    'fpga_type': 'xc3s500e-5-fg320',
    'flash_type': 'xcf04s',
    'project_name': 'textWriter'
})
config.read('text2spline.ini')
# -----------------------------------------------------------------------------
# ### for RAM init files
pre_text = config.get('RAM', 'pre_text')
text_input_file = config.get('RAM', 'text_input_file')
ram_x_file = config.get('RAM', 'ram_x_file')
ram_y_file = config.get('RAM', 'ram_y_file')
signal_bit_nb = config.getint('RAM', 'signal_bit_nb')
# -----------------------------------------------------------------------------
# ### for AMBA configuration files
generate_amba_files = config.getboolean('AMBA', 'generate_amba_files')
# -----------------------------------------------------------------------------
# ### for VHDL settings
update_period = config.getint('VHDL', 'update_period')
# -----------------------------------------------------------------------------
# ### for FPGA configuration files
generate_bitfile = config.getboolean('FPGA', 'generate_bitfile')
program_fpga = config.getboolean('FPGA', 'program_fpga')
program_flash = config.getboolean('FPGA', 'program_flash')
concat_file = config.get('FPGA', 'concat_file')
ucf_file = config.get('FPGA', 'ucf_file')
ise_work_path = config.get('FPGA', 'ise_work_path')
if not os.path.exists(ise_work_path):
    os.makedirs(ise_work_path)
ise_path = config.get('FPGA', 'ise_path')
fpga_type = config.get('FPGA', 'fpga_type')
flash_type = config.get('FPGA', 'flash_type')
# -----------------------------------------------------------------------------
# ### for general use
# used to name the output files and ISE temp files
project_name = config.get('GENERAL', 'project_name')
# =============================================================================
# ## internal global parameters
(output_path, dummy) = os.path.split(text_input_file)
output_path = os.path.join(output_path, 'export')
if not os.path.exists(output_path):
    os.makedirs(output_path)
(ram_x_file_path, ram_x_file_name) = os.path.split(ram_x_file)
(ram_y_file_path, ram_y_file_name) = os.path.split(ram_y_file)
ise_exec_path = os.path.join(ise_path, 'ISE\\bin\\nt64\\')
ise_work_file = os.path.join(ise_work_path, project_name)
output_text = []
text_list = []
# =============================================================================
# ## get text from csv
with open(text_input_file, 'rt') as csv_file:
    csv_object = csv.reader(csv_file)
    for row in csv_object:
        # remove leading/trailing whitespace
        row = [element.strip() for element in row]
        if not row[0].strip().startswith('#'):
            text_list.append(row)
for text in text_list:
    text.insert(0, pre_text)
# #############################################################################
# # generate RAM files and FPGA programming files for each text
for text in text_list:
    # =========================================================================
    # ## points of the figure to be drawn
    (x, y) = s_f.text2points(text)
    address_bit_nb = len(x).bit_length()
    point_nb = len(x)
    # -------------------------------------------------------------------------
    # ### spline interpolation
    xSpline = s_f.spline(x, oversampling_ratio=16, polynom_order=3)
    ySpline = s_f.spline(y, oversampling_ratio=16, polynom_order=3)
    # -------------------------------------------------------------------------
    # ### spline scaling
    x_scale = s_f.scale(x, xSpline)
    y_scale = s_f.scale(y, ySpline)
    xSpline_scale = s_f.scale(xSpline, xSpline)
    ySpline_scale = s_f.scale(ySpline, ySpline)
    # -------------------------------------------------------------------------
    # ### scaling
    # -------------------------------------------------------------------------
    # ### plot
    fig = plt.figure(1+text_list.index(text))
    ax = fig.add_subplot(111)
    ax.plot(xSpline_scale, ySpline_scale, 'b')
    #ax.plot(x, y, 'd')
    ax.set_xlim([0, 1])
    ax.set_ylim([0, 1])
    ax.axis('equal')
    plt.show()
    # -------------------------------------------------------------------------
    # ### output
    output_file = os.path.join(
        output_path,
        '-'.join(text) +
        '__' + '{}for{}'.format(address_bit_nb, point_nb) +
        '__' + project_name)
    output_text.append(output_file)
    # =========================================================================
    # ## scale to signal_bit_nb
    x_scale_2 = ((2**signal_bit_nb-1) * (x_scale-0.5)).round()
    x_scale_2[x_scale_2 < 0] = x_scale_2[x_scale_2 < 0] + 2**signal_bit_nb
    y_scale_2 = ((2**signal_bit_nb-1) * (y_scale-0.5)).round()
    y_scale_2[y_scale_2 < 0] = y_scale_2[y_scale_2 < 0] + 2**signal_bit_nb
    ram_x_file_id = open(ram_x_file, 'w')
    # =========================================================================
    # ## write RAM content to init and AMBA files
    if generate_amba_files:
        amba_file_id = open(output_file + '.txt', 'w')
        amba_binary_file_id = open(output_file + '.dat', 'wb')
        s_f.write_txt_bin('55 55\n',
                          amba_file_id, amba_binary_file_id)
        s_f.write_txt_bin('C0 80000300 00000002\n',
                          amba_file_id, amba_binary_file_id)
        s_f.write_txt_bin('55 55\n',
                          amba_file_id, amba_binary_file_id)
        s_f.write_txt_bin('55 55\n',
                          amba_file_id, amba_binary_file_id)
    for index in range(0, point_nb):
        string_format = '{0:0' + '{}'.format(signal_bit_nb) + 'b}\n'
        ram_x_file_id.write(string_format.format(int(x_scale_2[index])))
        if generate_amba_files:
            s_f.write_txt_bin(
                'C0 80000302 {0:08x}\n'.format(int(x_scale_2[index])).upper(),
                amba_file_id, amba_binary_file_id)
    for index in range(point_nb, 2**address_bit_nb-1):
        ram_x_file_id.write(('01'*(int(signal_bit_nb/2))) + '\n')
    ram_x_file_id.write(('01'*(int(signal_bit_nb/2))))
    ram_x_file_id.close()
    if generate_amba_files:
        s_f.write_txt_bin('55 55\n',
                          amba_file_id, amba_binary_file_id)
    ram_y_file_id = open(ram_y_file, 'w')
    for index in range(0, point_nb):
        string_format = '{0:0' + '{}'.format(signal_bit_nb) + 'b}\n'
        ram_y_file_id.write(string_format.format(int(y_scale_2[index])))
        if generate_amba_files:
            s_f.write_txt_bin(
                'C0 80000303 {0:08x}\n'.format(int(y_scale_2[index])).upper(),
                amba_file_id, amba_binary_file_id)
    for index in range(point_nb, 2**address_bit_nb-1):
        ram_y_file_id.write(('10'*(int(signal_bit_nb/2))) + '\n')
    ram_y_file_id.write(('10'*(int(signal_bit_nb/2))))
    ram_y_file_id.close()
    if generate_amba_files:
        s_f.write_txt_bin('55 55\n',
                          amba_file_id, amba_binary_file_id)
        s_f.write_txt_bin('C0 80000300 00{0:04X}01\n'.format(point_nb),
                          amba_file_id, amba_binary_file_id)
        s_f.write_txt_bin('C0 80000301 00000010\n',
                          amba_file_id, amba_binary_file_id)
        amba_binary_file_id.close()
        amba_file_id.close()
    # Copy init files
    shutil.copy(ram_x_file, output_file + '__' + ram_x_file_name)
    shutil.copy(ram_y_file, output_file + '__' + ram_y_file_name)
    # =========================================================================
    # ## initialize ISE
    ise = ISE.ISE(exec_path=ise_path,
                  work_path=ise_work_path,
                  source=concat_file,
                  ucf=ucf_file,
                  fpga=fpga_type,
                  flash=flash_type,
                  project_name=project_name)
    # =========================================================================
    # ## write config to concatenated.vhd
    if generate_bitfile:
        # Create temp file
        (temp_filedescriptor, temp_filename) = tempfile.mkstemp()
        temp_file_id = open(temp_filename, 'w')
        concat_file_id = open(concat_file, 'r')
        for line in concat_file_id:
            # remove comments
            line = re.sub(r"--.*", "", line.strip())
            # jump over empty lines
            print('line: {}'.format(line))
            if not line:
                continue
            # set address bit nb
            if 'constant patternAddressBitNb' in line:
                line = re.sub(r":=.*",
                              ":= {};".format(address_bit_nb),
                              line)
            # set point nb
            if 'constant curvePointNb' in line:
                line = re.sub(r":=.*",
                              ":= {};".format(point_nb),
                              line)
            # set update rate
            if 'updatePeriod <= to_unsigned(' in line:
                line = re.sub(r"\(.*",
                              "({}, updatePeriod'length);".format(update_period),
                              line)

            temp_file_id.write(line + '\n')
        temp_file_id.close()
        os.close(temp_filedescriptor)
        concat_file_id.close()
        # Remove original file
        os.remove(concat_file)
        # Move temp file to original file
        shutil.move(temp_filename, concat_file)
        # =====================================================================
        # ## synthesize, PAR and genereate bitfile
        shutil.copy(ise.generate_bitfile(), output_file + '.bit')
        # =====================================================================
        # ## generate mcs
        shutil.copy(ise.generate_mcs(), output_file + '.mcs')
    # =====================================================================
    # ## configure FPGA
    if program_fpga or program_flash:
        ise.configure_hw(program_flash, program_fpga)
# -----------------------------------------------------------------------------
print('#'*80)
print('GENERATED:')
print(' RAM files')
if generate_amba_files:
    print(' and AMBA files')
if generate_bitfile:
    print(' and FPGA files')
if program_flash:
    print('PROGRAMMED:')
    print(' FLASH')
    if program_fpga:
        print(' and FPGA')
else:
    if program_fpga:
        print('PROGRAMMED:')
        print(' FPGA')
print('\n'.join(output_text))
print('#'*80)
